import typing
from ..base import ApiResult, ApiMagic, magic_method
from ..common.account import (CurrencyInfo, NormalWallet, VirtualAccountInfo,
                              MobileInfo, BankInfo, EmailInfo, FeeRate,
                              VirtualAccount)


class BtcDepositAddress(ApiResult):
    walletAddress: str


class Balance(ApiResult):
    btc: CurrencyInfo[float]
    krw: CurrencyInfo[int]
    normalWallets: typing.List[NormalWallet]


class DailyBalanceInfo(ApiMagic):
    timestamp: int
    value: int
    krw: int
    btc: float


class DailyBalance(ApiResult):
    dailyBalance: typing.List[DailyBalanceInfo]


class _UserInfo(ApiMagic):
    virtualAccountInfo: VirtualAccountInfo
    mobileInfo: MobileInfo
    bankInfo: BankInfo
    emailInfo: EmailInfo
    securityLevel: int
    feeRate: FeeRate


class UserInfo(ApiResult):
    userInfo: _UserInfo


class Account(ApiMagic):
    @magic_method
    def btc_deposit_address(self) -> BtcDepositAddress:
        ...

    @magic_method
    def balance(self) -> Balance:
        ...

    @magic_method
    def daily_balance(self) -> DailyBalance:
        ...

    @magic_method
    def user_info(self) -> UserInfo:
        ...

    @magic_method
    def virtual_account(self) -> VirtualAccount:
        ...



import typing
from ..base import ApiResult, ApiMagic


class KrwHistoryEntry(ApiMagic):
    bankCode: int
    accountNumber: str
    depositor: str
    amount: int
    processLevel: int
    timestamp: int


class KrwHistory(ApiResult):
    krwHistory: typing.List[KrwHistoryEntry]



import typing
from ..base import ApiResult, ApiMagic, magic_method, post, form_data
from ..common.transaction import KrwHistory


class BtcHistoryEntry(ApiMagic):
    transactionId: str
    type: str
    from_: str
    to: str
    confirmations: int
    quantity: float
    timestamp: str


class BtcHistory(ApiResult):
    btcHistory: typing.List[BtcHistoryEntry]


class Transaction(ApiMagic):
    @magic_method
    def auth_number(self, type: str) -> ApiResult:
        ...

    @magic_method
    def btc_get(self) -> BtcHistory:
        ...

    @magic_method
    def krw_get(self) -> KrwHistory:
        ...

    @form_data
    @magic_method
    def btc_post(self, address: str, auth_number: int, qty: float,
                 type: str, from_address: str) -> ApiResult:
        ...



import typing
import functools
from .common.enums import Currency
from ..core import to_html


_T = typing.TypeVar('_T')


class ApiMagic:
    __origin__: typing.Type
    __args__: typing.Sequence[typing.Type]
    __annotations__: typing.Dict[str, typing.Type] = {}
    _subs: typing.Dict[typing.Any, typing.Type] = {}

    _annotations: typing.Dict[str, typing.Type] = {}

    @classmethod
    def _substitute(cls) -> typing.Dict[typing.Any, typing.Type]:
        if hasattr(cls, '__origin__') and hasattr(cls, '__args__'):
            subs = dict(cls.__origin__._subs)
            subs.update({p: a for p, a in zip(cls.__origin__.__parameters__,
                                              cls.__args__)})
            return subs
        return {}

    @classmethod
    def __init_subclass__(cls, *args: typing.Any, **kwargs: typing.Any)\
            -> None:
        cls._annotations = dict((k, v) for base in cls.__bases__
                                if hasattr(base, '__annotations__')
                                for k, v in base.__annotations__.items())
        cls._annotations.update(cls.__annotations__)

    def __init__(self, obj: typing.Any) -> None:
        self._obj = obj

    def __getattr__(self, key: str) -> typing.Any:
        obj = getattr(self._obj, key)
        if key in self._annotations:
            tp = self._annotations[key]
            tp = self._subs.get(tp, tp)
            if hasattr(tp, '__origin__') and\
                    getattr(tp, '__origin__') is typing.List:
                return [getattr(tp, '__args__')[0](elem) for elem in obj]
            if issubclass(tp, ApiMagic):
                subs = tp._substitute()
                res = subs.get(tp, tp)(obj)
                res._subs = subs
            else:
                res = tp(obj)
            return res
        else:
            return obj

    def _repr_html_(self) -> str:
        return to_html({
            key: getattr(self, key)
            for key in self._annotations if not key.startswith('_')
        })


class CurrencyMagic(ApiMagic):
    _currency = Currency.BTC

    def __getitem__(self: _T, currency: Currency) -> _T:
        return _set_currency(self, currency)


class ApiResult(ApiMagic):
    result: str
    errorCode: str


_T_magic = typing.TypeVar('_T_magic', bound=ApiMagic, covariant=True)


def magic_method(fn: _T) -> _T:
    fn_ = typing.cast(typing.Callable[..., typing.Any], fn)
    ret_type: ApiMagic = fn_.__annotations__['return']
    name = fn_.__name__
    def_kwargs = {}

    prefix, *postfix = name.rsplit('_', 2)
    if postfix and postfix[0] in ('get', 'post'):
        def_kwargs['_method_'] = postfix[0]
        name = prefix

    @functools.wraps(fn_)
    def wrapper(self: typing.Any, *args: typing.Any, **kwargs: typing.Any)\
            -> _T_magic:
        kwargs_ = dict(def_kwargs)
        kwargs_.update(kwargs)
        if hasattr(self, '__getattr__'):
            obj: typing.Any = self.__getattr__(name)(*args, **kwargs_)
            return typing.cast(_T_magic, ret_type(obj))
        else:
            raise AttributeError(name)

    return typing.cast(_T, wrapper)


def _add_param(fn: _T, **kwargs: typing.Any) -> _T:
    fn_ = typing.cast(typing.Callable[..., typing.Any], fn)

    @functools.wraps(fn_)
    def wrapper(*args: typing.Any, **kwargs_: typing.Any) -> typing.Any:
        p = dict(kwargs)
        p.update(kwargs_)

        return fn_(*args, **p)

    return typing.cast(_T, wrapper)


def delegated_getter(name: str) -> typing.Callable[[_T], _T]:
    def decorator(fn: _T) -> _T:
        fn_ = typing.cast(typing.Callable, fn)

        @functools.wraps(fn_)
        def wrapper(self: typing.Any) -> typing.Any:
            return getattr(getattr(self, name), fn_.__name__)

        return typing.cast(_T, wrapper)
    return decorator


def delegated(name: str) -> typing.Callable[[_T], _T]:
    def decorator(fn: _T) -> _T:
        fn_ = typing.cast(typing.Callable, fn)

        @functools.wraps(fn_)
        def wrapper(self: typing.Any, *args: typing.Any,
                    **kwargs: typing.Any) -> typing.Any:
            return getattr(getattr(self, name), fn_.__name__)(*args, **kwargs)

        return typing.cast(_T, wrapper)
    return decorator


def magic_result(fn: _T) -> _T:
    fn_ = typing.cast(typing.Callable, fn)
    ret_type: typing.Type = fn_.__annotations__['return']

    @functools.wraps(fn_)
    def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        return ret_type(fn_(*args, **kwargs))

    return typing.cast(_T, wrapper)


def post(fn: _T) -> _T:
    return _add_param(fn, _method_='post')


def form_data(fn: _T) -> _T:
    return _add_param(fn, _payload_='data')


def currency_method(fn: _T) -> _T:
    fn_ = typing.cast(typing.Callable, fn)

    @functools.wraps(fn_)
    def wrapper(self: ApiMagic, *args: typing.Any, **kwargs: typing.Any)\
            -> typing.Any:
        if hasattr(self, '_currency') and 'currency' not in kwargs:
            kwargs['currency'] = getattr(self, '_currency')
        return fn_(self, *args, **kwargs)

    return typing.cast(_T, wrapper)


def _set_currency(self: _T, currency: Currency) -> _T:
    if isinstance(self, CurrencyMagic):
        obj = typing.cast(_T, self.__class__(self._obj))
        if isinstance(obj, CurrencyMagic):
            obj._currency = currency
        return obj
    return self



import typing
from ..base import (ApiResult, CurrencyMagic, magic_method, post, form_data,
                    currency_method)
from ..common.order import OrderId, CompleteOrders, LimitOrders
from ..common.enums import Currency, OrderType


class Order(CurrencyMagic):
    @currency_method
    @magic_method
    def cancel(self, order_id: str, price: int, qty: float,
               is_ask: OrderType,
               currency: Currency=Currency.BTC) -> ApiResult:
        ...

    @currency_method
    @magic_method
    def limit_buy(self, price: int, qty: float,
                  currency: Currency=Currency.BTC) -> OrderId:
        ...

    @currency_method
    @magic_method
    def limit_sell(self, price: int, qty: float,
                   currency: Currency=Currency.BTC) -> OrderId:
        ...

    @currency_method
    @magic_method
    def complete_orders(self, currency: Currency=Currency.BTC)\
            -> CompleteOrders:
        ...

    @currency_method
    @magic_method
    def limit_orders(self, currency: Currency=Currency.BTC)\
            -> LimitOrders:
        ...



import typing
from ..base import ApiResult, ApiMagic


_T_currency = typing.TypeVar('_T_currency', int, float)


class CurrencyInfo(ApiMagic, typing.Generic[_T_currency]):
    avail: _T_currency
    balance: _T_currency


class NormalWallet(ApiMagic):
    balance: float
    label: str


class VirtualAccountInfo(ApiMagic):
    depositor: str
    accountNumber: str
    bankName: str


class MobileInfo(ApiMagic):
    userName: str
    phoneNumber: str
    phoneCorp: str
    isAuthenticated: bool


class BankInfo(ApiMagic):
    depositor: str
    bankCode: str
    accountNumber: str
    isAuthenticated: bool


class EmailInfo(ApiMagic):
    isAuthenticated: bool
    email: str


class FeeRate(ApiMagic):
    maker: float
    taker: float


class VirtualAccount(ApiResult):
    accountNumber: str
    depositor: str
    bankName: str



from . import raw, error
import typing
import time
import threading
import functools
import json
import collections
import html


def to_html(obj: typing.Any) -> str:
    if isinstance(obj, dict):
        head_row =\
                '<tr><th style="width: auto; white-space: nowrap">Field</th>'\
                '<th style="width: 100%">Value</th>'
        return "".join((
            elem for elems in [
                [
                    '<table style="table-layout: auto; width: 100%"><thead>',
                    head_row,
                    "</tr></thead><tbody>"
                ],
                (
                    '<tr><td><pre style="width: auto; white-space: nowrap">'
                    '{field}</pre></td>'
                    "<td>{value}</td></tr>".format(
                        field=key,
                        value=to_html(value)
                    ) for key, value in obj.items()
                ),
                [
                    "</tbody></table>"
                    '</div>'
                ]
            ] for elem in elems))
    elif isinstance(obj, list):
        return "".join(term for terms in [
            ['<ul style="max-height: 20em; overflow-y: auto">'],
            ("<li>" + to_html(elem) + "</li>" for elem in obj),
            ["</ul>"]
        ] for term in terms)
    elif isinstance(obj, bool):
        return "🔵 True" if obj else "🔴 False"
    elif hasattr(obj, '_repr_html_'):
        return str(obj._repr_html_())
    else:
        return ''.join([
            '<span onclick="$(\'input\', this).show().val('
            '$(\'span\', this).hide().text()).focus()[0].'
            'select(); document.execCommand(\'copy\')"><span>',
            html.escape(repr(obj)),
            '</span><input style="text-align: right; '
            'display: none; border: none; width: 100%; '
            'background: transparent;"></span>'])


def _process_error(path: str, response: typing.Mapping[str, typing.Any])\
        -> typing.Mapping[str, typing.Any]:
    if 'result' not in response:
        raise RuntimeError(f"Ugly-shaped result: {repr(response)}")

    if response['result'] != 'success':
        code = int(response['errorCode'])
        msg = response.get('errorMessage',
                           error.CoinoneErrorCode(code).__doc__)
        raise error.CoinoneError(code, f'{msg} at {path}')
    return response


_AnyFn = typing.Callable[..., typing.Any]


def _rate_limited(count: int, period: float)\
        -> typing.Callable[[_AnyFn], _AnyFn]:
    lock = threading.Lock()
    last_call_time: float = 0
    left_call_count = count

    def decorator(fn: _AnyFn) -> _AnyFn:
        @functools.wraps(fn)
        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            nonlocal last_call_time, left_call_count

            now = time.time()
            lock.acquire()
            if last_call_time < now - period or left_call_count <= 0:
                if left_call_count <= 0:
                    time_to_sleep = last_call_time + period - now
                    if time_to_sleep > 0:
                        time.sleep(time_to_sleep)
                last_call_time = time.time()
                left_call_count = count
            left_call_count -= 1
            lock.release()

            return fn(*args, **kwargs)

        return wrapper

    return decorator


class DictObject(dict):
    def __getattr__(self, key: str) -> typing.Any:
        key = key.rstrip('_')
        if key in self:
            return self[key]
        else:
            raise AttributeError(key)

    def __setattr__(self, key: str, value: typing.Any) -> None:
        self[key.rstrip('_')] = value

    def __delattr__(self, key: str) -> None:
        key = key.rstrip('_')
        if key in self:
            del self[key]
        else:
            raise AttributeError(key)

    def __dir__(self) -> typing.Iterable[str]:
        return [key for keys in (super().__dir__(), self.keys())
                for key in keys]

    def _repr_html_(self) -> str:
        return to_html(self)


class _Coinone(raw.Coinone):
    def _execute(self, path: str, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        json_opt = kwargs.get('_json_', {})
        json_opt['object_hook'] = DictObject
        kwargs['_json_'] = json_opt
        return _process_error(path, super()._execute(path, **kwargs))


class Coinone(_Coinone):
    @_rate_limited(90, 60)
    def _execute(self, path: str, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        return super()._execute(path, **kwargs)


class CoinoneV1(_Coinone, raw.CoinoneV1):
    @_rate_limited(6, 1)
    def _execute(self, path: str, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        return super()._execute(path, **kwargs)


class CoinoneV2(_Coinone, raw.CoinoneV2):
    @_rate_limited(6, 1)
    def _execute(self, path: str, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        return super()._execute(path, **kwargs)



import typing
from ..base import ApiResult, ApiMagic, magic_method
from ..common.account import (CurrencyInfo, NormalWallet, VirtualAccountInfo,
                              MobileInfo, BankInfo, EmailInfo, FeeRate,
                              VirtualAccount)


class BtcDepositAddress(ApiResult):
    walletAddress: str


class Balance(ApiResult):
    btc: CurrencyInfo[float]
    bch: CurrencyInfo[float]
    eth: CurrencyInfo[float]
    etc: CurrencyInfo[float]
    xrp: CurrencyInfo[float]
    qtum: CurrencyInfo[float]
    iota: CurrencyInfo[float]
    ltc: CurrencyInfo[float]
    btg: CurrencyInfo[float]
    krw: CurrencyInfo[int]
    normalWallets: typing.List[NormalWallet]


class DailyBalanceInfo(ApiMagic):
    timestamp: int
    value: int
    krw: int
    btc: float
    bch: float
    eth: float
    etc: float
    xrp: float
    qtum: float
    iota: float
    ltc: float
    btg: float


class DailyBalance(ApiResult):
    dailyBalance: typing.List[DailyBalanceInfo]


class WalletAddress(ApiMagic):
    btc: str
    bch: str
    eth: str
    etc: str
    xrp: str
    xrp_tag: str
    qtum: str
    iota: str
    ltc: str
    btg: str


class DepositAddress(ApiResult):
    walletAddress: WalletAddress


class FeeRates(ApiMagic):
    btc: FeeRate
    bch: FeeRate
    eth: FeeRate
    etc: FeeRate
    xrp: FeeRate
    qtum: FeeRate
    iota: FeeRate
    ltc: FeeRate
    btg: FeeRate


class _UserInfo(ApiMagic):
    virtualAccountInfo: VirtualAccountInfo
    mobileInfo: MobileInfo
    bankInfo: BankInfo
    emailInfo: EmailInfo
    securityLevel: int
    feeRate: FeeRates


class UserInfo(ApiResult):
    userInfo: _UserInfo


class Account(ApiMagic):
    @magic_method
    def btc_deposit_address(self) -> BtcDepositAddress:
        ...

    @magic_method
    def balance(self) -> Balance:
        ...

    @magic_method
    def daily_balance(self) -> DailyBalance:
        ...

    @magic_method
    def user_info(self) -> UserInfo:
        ...

    @magic_method
    def virtual_account(self) -> VirtualAccount:
        ...



import typing
from ..base import ApiResult, ApiMagic


class OrderId(ApiResult):
    orderId: str


class CompleteOrder(ApiMagic):
    timestamp: int
    price: int
    qty: float
    type: str
    feeRate: float
    fee: float
    orderId: str


class CompleteOrders(ApiResult):
    completeOrders: typing.List[CompleteOrder]


class LimitOrder(ApiMagic):
    index: int
    timestamp: int
    price: int
    qty: float
    orderId: str
    type: str
    feeRate: float


class LimitOrders(ApiResult):
    limitOrders: typing.List[LimitOrder]



import typing
from ..base import ApiResult, ApiMagic, magic_method, post, form_data
from ..common.order import OrderId, CompleteOrders, LimitOrders


class Order(ApiMagic):
    @post
    @form_data
    @magic_method
    def cancel(self, order_id: str, price: int, qty: float, is_ask: int)\
            -> ApiResult:
        ...

    @post
    @form_data
    @magic_method
    def limit_buy(self, price: int, qty: float) -> OrderId:
        ...

    @post
    @form_data
    @magic_method
    def limit_sell(self, price: int, qty: float) -> OrderId:
        ...

    @magic_method
    def complete_orders(self) -> CompleteOrders:
        ...

    @magic_method
    def limit_orders(self) -> LimitOrders:
        ...



from setuptools import setup

setup(
    name='pycoinone',
    version='2017.12.29a0',
    packages=[
        'coinone',
        'coinone.api',
        'coinone.api.common',
        'coinone.api.v1',
        'coinone.api.v2'
    ],
    url='http://github.com/gwangyi/pycoinone',
    license='MIT',
    author='Sungkwang Lee',
    author_email='gwangyi.kr@gmail.com',
    description='Python Coinone API wrapper',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6'
    ],
    install_requires=['requests']
)



import typing
from ..base import ApiResult, ApiMagic, magic_method, post, form_data
from ..common.transaction import KrwHistory


class CoinTransactionEntry(ApiMagic):
    txid: str
    type: str
    from_: str
    to: str
    confirmations: int
    quantity: float
    timestamp: int


class CoinTransactions(ApiResult):
    transactions: typing.List[CoinTransactionEntry]


class SendCoin(ApiResult):
    txid: str


class KrwMagic(ApiMagic):
    @magic_method
    def history(self) -> KrwHistory:
        ...


class Transaction(ApiMagic):
    @magic_method
    def auth_number(self, type: str) -> ApiResult:
        ...

    @magic_method
    def history(self, currency: str) -> CoinTransactions:
        ...

    @property
    def krw(self) -> KrwMagic:
        return KrwMagic(self.__getattr__('krw'))

    @magic_method
    def coin(self, currency: str, address: str, auth_number: int, qty: float,
             type: str, from_address: str) -> ApiResult:
        ...



import typing
from ..base import ApiResult, ApiMagic, magic_result, delegated
from ...core import Coinone


class OrderbookEntry(ApiMagic):
    price: int
    qty: float


class Orderbook(ApiResult):
    ask: typing.List[OrderbookEntry]
    bid: typing.List[OrderbookEntry]
    timestamp: int
    currency: str


class TradesEntry(ApiMagic):
    price: int
    qty: float
    timestamp: int


class Trades(ApiResult):
    timestamp: int
    completeOrders: typing.List[TradesEntry]
    currency: str


class _Ticker(ApiMagic):
    high: int
    low: int
    last: int
    first: int
    volume: float
    yesterday_high: int
    yesterday_low: int
    yesterday_last: int
    yesterday_first: int
    yesterday_volume: float
    currency: str


class Ticker(_Ticker, ApiResult):
    timestamp: int


class TickerAll(ApiResult):
    btc: _Ticker
    bch: _Ticker
    eth: _Ticker
    etc: _Ticker
    xrp: _Ticker
    qtum: _Ticker
    iota: _Ticker
    ltc: _Ticker
    btg: _Ticker
    timestamp: int


class PublicApiMixin:
    def __init__(self) -> None:
        self._public_api = Coinone()

    @magic_result
    @delegated('_public_api')
    def orderbook(self, currency: str) -> Orderbook:
        ...

    @magic_result
    @delegated('_public_api')
    def trades(self, currency: str, period: str) -> Trades:
        ...

    @magic_result
    @delegated('_public_api')
    def ticker(self, currency: str) -> Ticker:
        ...

    def ticker_all(self) -> TickerAll:
        return TickerAll(self._public_api.ticker(currency='all'))



import typing

from .account import Account
from .order import Order
from .transaction import Transaction
from ..base import magic_result, delegated_getter
from ..common.public import PublicApiMixin
from ...core import CoinoneV1 as Api, Coinone


class CoinoneV1(PublicApiMixin):
    def __init__(self, access_token: str) -> None:
        super().__init__()
        self._api = Api(access_token)

    @property
    def account(self) -> Account:
        return Account(self._api.account)

    @property
    def order(self) -> Order:
        return Order(self._api.order)

    @property
    def transaction(self) -> Transaction:
        return Transaction(self._api.transaction)



import enum


class Currency(str, enum.Enum):
    BTC = 'btc'
    BCH = 'bch'
    ETH = 'eth'
    ETC = 'etc'
    XRP = 'xrp'
    QTUM = 'qtum'
    IOTA = 'iota'
    LTC = 'ltc'
    BTG = 'btg'

    KRW = 'krw'


class OrderType(int, enum.Enum):
    Ask = 1
    Bid = 0
    Sell = 1
    Buy = 0



import json
import time
import base64
import hashlib
import hmac
import typing
import requests
import sys


def _default_nonce() -> int:
    return int(time.time() * 1000)


class Coinone:
    _base_url = 'https://api.coinone.co.kr/'
    _preferred_method = 'get'
    _preferred_payload = 'params'

    @staticmethod
    def _kwargs_to_payload(kwargs: typing.Mapping[str, typing.Any])\
            -> typing.Mapping[str, typing.Any]:
        payload = {k: v for k, v in kwargs.items()
                   if k[:1] != '_' and k[-1:] != '_'}
        return payload

    def __getattr__(self, key: str) -> '_APIEndpoint':
        return _APIEndpoint(self, '/' + key)

    def _execute(self, path: str, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        url = self._base_url + path.lstrip('/')
        headers = kwargs.get('_headers_', {})
        method = kwargs.get('_method_', self._preferred_method).upper()
        json_opt = kwargs.get('_json_', {})
        payload = self._kwargs_to_payload(kwargs)

        payload_type = kwargs.get('_payload_', self._preferred_payload)
        r = requests.request(method, url, headers=headers,
                             **{payload_type: payload})

        try:
            return typing.cast(typing.Mapping[str, typing.Any],
                               r.json(**json_opt))
        except Exception:
            sys.stderr.write(r.text)
            raise


class CoinoneV1(Coinone):
    _base_url = 'https://api.coinone.co.kr/v1/'

    def __init__(self, access_token: str) -> None:
        super().__init__()
        self._access_token = access_token

    def _execute(self, path: str, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        if 'access_token' in kwargs:
            del kwargs['access_token']
        return super()._execute(path, access_token=self._access_token,
                                **kwargs)


class CoinoneV2(Coinone):
    _base_url = 'https://api.coinone.co.kr/v2/'
    _preferred_method = 'post'
    _preferred_payload = 'json'

    def __init__(self, access_token: str, secret_key: str,
                 nonce: typing.Callable[[], typing.Any] = _default_nonce)\
            -> None:
        super().__init__()
        self._nonce = nonce
        self._access_token = access_token
        self._secret_key = secret_key

    def _encode_payload(self, payload: typing.Mapping[str, typing.Any])\
            -> bytes:
        dumped_json = json.dumps(payload)
        return base64.b64encode(dumped_json.encode('utf-8'))

    def _signature(self, payload: bytes) -> str:
        signature = hmac.new(self._secret_key.upper().encode('ascii'),
                             payload,
                             hashlib.sha512)
        return signature.hexdigest()

    def _execute(self, path: str, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        headers: typing.Dict[str, str] = kwargs.get('_headers_', {})
        kwargs['access_token'] = self._access_token
        kwargs['nonce'] = self._nonce()
        payload = self._kwargs_to_payload(kwargs)

        encoded_payload: bytes = self._encode_payload(payload)

        headers.update({
            'X-COINONE-PAYLOAD': encoded_payload.decode('ascii'),
            'X-COINONE-SIGNATURE': self._signature(encoded_payload)
        })
        if '_headers_' in kwargs:
            del kwargs['_headers_']

        return super()._execute(path, _headers_=headers, **kwargs)


class _APIEndpoint:
    def __init__(self, session: Coinone, name: str) -> None:
        self._session = session
        self._name = name

    def __call__(self, **kwargs: typing.Any)\
            -> typing.Mapping[str, typing.Any]:
        return self._session._execute(self._name, **kwargs)

    def __getattr__(self, key: str) -> '_APIEndpoint':
        return _APIEndpoint(self._session, '/'.join((self._name, key)))



import enum
import typing
import collections


class CoinoneErrorCode(enum.Enum):
    BlockedUserAccess = (4, 'Blocked user access')
    AccessTokenMissing = (11, 'Access token is missing')
    InvalidAccessToken = (12, 'Invalid access token')
    PermissionDenied = (40, 'Invalid API permission')
    AuthenticateError = (50, 'Authenticate error')
    InvalidAPI = (51, 'Invalid API')
    DeprecatedAPI = (52, 'Deprecated API')
    TwoFactorAuthFail = (53, 'Two Factor Auth Fail')
    SessionExpired = (100, 'Session expired')
    InvalidFormat = (101, 'Invalid format')
    IdNotExist = (102, "ID is not exist")
    LackOfBalance = (103, "Lack of Balance")
    OrderIdNotExist104 = (104, "Order id is not exist")
    PriceIsNotCorrect = (105, "Price is not correct")
    LockingError106 = (106, "Locking error")
    ParameterError = (107, "Parameter error")
    OrderIdNotExist111 = (111, "Order id is not exist")
    CancelFailed = (112, "Cancel failed")
    TooLowQuantity = (113, "Quantity is too low(ETH, ETC > 0.01)")
    V2PayloadMissing = (120, "V2 API payload is missing")
    V2SignatureMissing = (121, "V2 API signature is missing")
    V2NonceMissing = (122, "V2 API nonce is missing")
    V2SignatureMismatch = (123, "V2 API signature is not correct")
    V2NonceNegative = (130, "V2 API Nonce value must be a positive integer")
    V2NonceMismatch = (131, "V2 API Nonce is must be bigger then last nonce")
    V2BodyCorrupted = (132, "V2 API body is corrupted")
    TooManyLimitOrders = (141, "Too many limit orders")
    V1WrongAccessToken = (150, "It's V1 API. "
                               "V2 Access token is not acceptable")
    V2WrongAccessToken = (151, "It's V2 API. "
                               "V1 Access token is not acceptable")
    WalletError = (200, "Wallet Error")
    LimitationError202 = (202, "Limitation error")
    LimitationError210 = (210, "Limitation error")
    LimitationError220 = (220, "Limitation error")
    LimitationError221 = (221, "Limitation error")
    MobileAuthError310 = (310, "Mobile auth error")
    NeedMobileAuth = (311, "Need mobile auth")
    BadName = (312, "Name is not correct")
    BadPhoneNumber330 = (330, "Phone number error")
    PageNotFound = (404, "Page not found error")
    ServerError = (405, "Server error")
    LockingError444 = (444, "Locking error")
    EmailError500 = (500, "Email error")
    EmailError501 = (501, "Email error")
    MobileAuthError777 = (777, "Mobile auth error")
    BadPhoneNumber778 = (778, "Phone number error")
    AppNotFound = (1202, "App not found")
    AlreadyRegistered = (1203, "Already registered")
    InvalidAccess = (1204, "Invalid access")
    APIKeyError = (1205, "API Key error")
    UserNotFound1206 = (1206, "User not found")
    UserNotFound1207 = (1207, "User not found")
    UserNotFound1208 = (1208, "User not found")
    UserNotFound1209 = (1209, "User not found")

    @typing.no_type_check
    def __new__(cls, code: int, desc: str) -> 'CoinoneErrorCode':
        self = object.__new__(cls)
        self._value_ = code
        self.__doc__ = desc
        return self

    def __init__(cls, code: int, desc: typing.Optional[str] = None) -> None:
        pass

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}.{self.name}: '\
               f'{self.__doc__} ({self.value})>'


class CoinoneError(RuntimeError):
    _error_map: typing.Dict[typing.Container[int],
                            typing.Type['CoinoneError']]\
            = collections.OrderedDict()

    @staticmethod
    def register(*error_code: CoinoneErrorCode)\
            -> typing.Callable[[typing.Type['CoinoneError']],
                               typing.Type['CoinoneError']]:
        def decorator(error_type: typing.Type['CoinoneError'])\
                -> typing.Type['CoinoneError']:
            codes = tuple(code.value for code in error_code)
            CoinoneError._error_map[codes] = error_type
            return error_type

        return decorator

    def __new__(cls, code: int, message: typing.Optional[str]=None)\
            -> 'CoinoneError':
        if cls is CoinoneError:
            for k, v in CoinoneError._error_map.items():
                if code in k:
                    return v(code, message)
        return typing.cast(CoinoneError, RuntimeError.__new__(cls))

    def __init__(self, code: int, message: typing.Optional[str]=None) -> None:
        ecode = CoinoneErrorCode(code)
        if message is None:
            super().__init__(f'[Error {code}] {ecode.__doc__}')
        else:
            super().__init__(f'[Error {code}] {message}')
        self.code = code


@CoinoneError.register(CoinoneErrorCode.APIKeyError)
class APIKeyErrorError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.AccessTokenMissing)
class AccessTokenMissingError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.AlreadyRegistered)
class AlreadyRegisteredError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.AppNotFound)
class AppNotFoundError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.AuthenticateError)
class AuthenticateErrorError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.BadName)
class BadNameError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.BadPhoneNumber330,
                       CoinoneErrorCode.BadPhoneNumber778)
class BadPhoneNumberError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.BlockedUserAccess)
class BlockedUserAccessError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.CancelFailed)
class CancelFailedError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.DeprecatedAPI)
class DeprecatedAPIError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.EmailError500,
                       CoinoneErrorCode.EmailError501)
class EmailErrorError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.IdNotExist)
class IdNotExistError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.InvalidAPI)
class InvalidAPIError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.InvalidAccess)
class InvalidAccessError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.InvalidAccessToken)
class InvalidAccessTokenError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.InvalidFormat)
class InvalidFormatError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.LackOfBalance)
class LackOfBalanceError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.LimitationError202,
                       CoinoneErrorCode.LimitationError210,
                       CoinoneErrorCode.LimitationError220,
                       CoinoneErrorCode.LimitationError221)
class LimitationError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.LockingError106,
                       CoinoneErrorCode.LockingError444)
class LockingError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.MobileAuthError310,
                       CoinoneErrorCode.MobileAuthError777)
class MobileAuthErrorError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.NeedMobileAuth)
class NeedMobileAuthError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.OrderIdNotExist104,
                       CoinoneErrorCode.OrderIdNotExist111)
class OrderIdNotExistError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.PageNotFound)
class PageNotFoundError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.ParameterError)
class ParameterErrorError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.PermissionDenied)
class PermissionDeniedError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.PriceIsNotCorrect)
class PriceIsNotCorrectError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.ServerError)
class ServerErrorError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.SessionExpired)
class SessionExpiredError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.TooLowQuantity)
class TooLowQuantityError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.TooManyLimitOrders)
class TooManyLimitOrdersError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.TwoFactorAuthFail)
class TwoFactorAuthFailError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.UserNotFound1206,
                       CoinoneErrorCode.UserNotFound1207,
                       CoinoneErrorCode.UserNotFound1208,
                       CoinoneErrorCode.UserNotFound1209)
class UserNotFoundError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V1WrongAccessToken)
class V1WrongAccessTokenError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2BodyCorrupted)
class V2BodyCorruptedError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2NonceMismatch)
class V2NonceMismatchError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2NonceMissing)
class V2NonceMissingError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2NonceNegative)
class V2NonceNegativeError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2PayloadMissing)
class V2PayloadMissingError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2SignatureMismatch)
class V2SignatureMismatchError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2SignatureMissing)
class V2SignatureMissingError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.V2WrongAccessToken)
class V2WrongAccessTokenError(CoinoneError):
    pass


@CoinoneError.register(CoinoneErrorCode.WalletError)
class WalletErrorError(CoinoneError):
    pass
